from .onedimprt_classifier import * 

"""
A memory-intensive extension to classifier<RecursiveOneDimClassifier>. 
Practically memorizes all (index,label) pairs for label vector L. 
With this information, classifier uses its root node to add the appropriate 
vector v_l to every input v, before classifier<RecursiveOneDimClassifier> 
proceeds with classifying (v_l + v). 

Due to this classifier's additive adjustment scheme relying on indices, 
the ordering of input samples matter for accuracy. In other words, a 
permutation D_ of the samples of dataset D will produce different output 
labels. This is a fundamental weakness of this memory-based schematic. 
"""
class AdditiveAdjustedRecursiveODC(RecursiveOneDimClassifier): 

    def __init__(self,D,L,prg=None,pscheme=0,verbose:bool=False): 
        D = self.preproc(D,L)
        super().__init__(D,L,prg,pscheme,verbose) 

    #--------------------------------------------------------------------

    def preproc(self,D,L): 
        self.make_additive_map(D,L) 
        self.make_adjustment(L) 
        self.root_ = ODCAdditiveAdjustmentNode(self.crepr,self.add_map) 

        D2 = np.array([self.root_.map(d) for d in D]) 
        self.root_.reset() 
        return D2 

    def make_adjustment(self,L): 
        self.crepr = indexed_contiguous_repr__sequence(L)

        
    def make_additive_map(self,D,L):
        self.add_map = default_additive_map(D,L) 

    #----------------------------------------------------------------------

    """
    main method #2 
    """
    def classify(self,v): 
        v2 = self.root_.map(v) 
        return super().classify(v2) 

    """
    main method #3 
    """
    # NOTE: duplicate code from <OneDimClassifier> 
    """
    def score_accuracy(self,D,L): 
        assert len(D) == len(L) 

        c = 0 
        for (x,l) in zip(D,L): 
            l_ = self.classify(x) 
            if l_ == l: 
                c += 1 
        return c 
    """ 

#------------------------------------------------------------------------------------------------


"""
Memory-Bound Alternating Additive Adjusted 
Recursive One-Dimensional Classifier 

This classifier works on the same principle as classifier<AdditiveAdjustedRecursiveODC>. 
However, it has additional parameters (`max_mem_sz`,`mem_ratio_per_node`) that limit the 
number of index-to-label ranges it can recall. 

Unlike classifier<AdditiveAdjustedRecursiveODC>, a structure that sets one node for additive adjustment
as the root, and then proceeds to classify the offsetted input vectors according to 
classifier<RecursiveOneDimClassifier>, every node of classifier<MBAAARecursiveODC> is of 
class<ODCAAPrefixedNode>. For an input vector v_i, this node adds a vector v_il, according to 
its index-to-label range information, to v_i, and then uses classifier<AdditiveAdjustedRecursiveODC> 
to classify (v_il + v_i). 

For sub-data D_ of D, and its associate index-to-label ranges R_, class<ODCAAPrefixedNode> can 
memorize at most 
    min([ceil(`mem_ratio_per_node` * |R|),`max_mem_sz` - Q]); Q the number of ranges memorized by previous 
            nodes to the node instance. 
"""
class MBAAARecursiveODC(RecursiveOneDimClassifier): 

    def __init__(self,D,L,prg=None,pscheme=0,max_mem_sz=100,mem_ratio_per_node=0.5,\
        verbose:bool=False): 

        assert type(max_mem_sz) == int and max_mem_sz > 0 
        assert 0.0 < mem_ratio_per_node <= 1.0

        super().__init__(D,L,prg,pscheme,verbose) 


        self.max_mem_sz = max_mem_sz
        self.current_mem_sz = 0 
        self.mem_ratio_per_node = mem_ratio_per_node

        # every element is 
        # ODCNode
        # P := dict, label -> {indices of label in sub-data}
        # E := dict, label -> number of positive classification errors in D  
        # D := 2-d matrix, sub-data 
        # L := vector, labels corresponding to sub-data 
        return

    def init_root(self):
        self.root = ODCAAPrefixedNode(self.D,self.L,previous_indices=set()) 
        self.node_cache.append(self.root) 
        return

    def __next__(self): 
        if len(self.node_cache) == 0: 
            self.fin_stat = True 
        
        if self.fin_stat: 
            return False 

        node = self.node_cache.pop(0) 
        sz,child_nodes = node.make(self.max_mem_sz - self.current_mem_sz,\
            self.mem_ratio_per_node,self.prg,pscheme=self.pscheme)
        self.current_mem_sz += sz 
        self.node_cache.extend(child_nodes)

    """
    main method #1
    """
    def fit(self): 
        self.init_root() 
        while not self.fin_stat: 
            next(self) 
        self.clear_data() 

    """
    main method #2 
    """
    def classify(self,v): 
        q = self.root 
        l = None 
        self.sample_cls_path.clear() 
        while type(q) != type(None): 
            v,l,q = q.classify(v) 
            self.sample_cls_path.append((l,q))  
        return l 

    def clear_data(self): 
        self.D,self.L = None,None 


from .numerical_generator import * 
from .search_space_iterator import * 
from types import MethodType,FunctionType
from copy import deepcopy

class XClassifier: 

    def __init__(self,D,L,index): 
        assert is_2dmatrix(D) 
        assert is_vector(L) 
        assert 0 <= index < D.shape[1] 

        self.D = D 
        self.D_ = deepcopy(D) 
        self.index = index 
        self.L = L 
        self.label2index_map = dict() 

    def partition_eval(self): 
        l_info = dict() 
        q = set(self.L) 

        for q_ in q: 
            l_info[q_] = list(self.index_info(q_))
            
        return l_info 

    """
    return: 
    - (min value,max value, mean value) of label along `index`
    """
    def index_info(self,l,store_indices:bool=True): 
        if l not in self.label2index_map: 
            indices = self.indices_of_label(l)
            assert len(indices) > 0 
            if store_indices: 
                self.label2index_map[l] = indices 
        else: 
            indices = self.label2index_map[l] 

        values = sorted([self.D_[i,self.index] for i in indices])
        min,max = values[0],values[-1] 
        ave = np.mean(values)
        return min,max,ave 

    def indices_of_label(self,l): 
        return np.where(self.L==l)[0] 

#---------------------------------------------------------------------------------

def cumulative_vector_difference(V,add_log:bool=False): 
    S = 0 
    for i in range(len(V) - 1): 
        q = V[i] 
        for j in range(i+1,len(V)): 
            q2 = abs(q-V[j]) 
            if add_log: 
                q2 = abs(safe_log(q2,0)) 
            S += q2 
    return S 

# NOTE: careful with large vector values 
"""
used by class<ModDivClassifier>. Finds weight vector `best_W`, comprised 
of {-1,10} that maximizes differences between samples `s_i x best_W`. 
"""
class OneTenLinearFunctionDifferenceMaximizer:

    def __init__(self,samples,alter_pattern="linear",verbose=False):
        assert is_2dmatrix(samples) 
        assert alter_pattern in {"linear","combinative"}  

        self.samples = samples 
        self.alter_pattern = alter_pattern 
        self.verbose = verbose 

        self.W = -np.ones((self.samples.shape[1],)) 
        self.index = 0 

        self.best_W = None 
        self.best_score = -float('inf') 

        q = np.zeros((self.samples.shape[1],2))
        #q[:,0],q[:,1] = 1,19 
        q[:,0],q[:,1] = -1,3 

        self.set_index_ordering()
        self.ssi = SearchSpaceIterator(q,deepcopy(self.W),\
            self.ordering,SSIHop = 2,cycleOn = False, cycleIs = 0) 
        self.fin_stat = False 
        return 

    def set_index_ordering(self): 
        diff_vec = [(i,cumulative_vector_difference(self.samples[:,i],True)) for i in range(self.samples.shape[1])]
        diff_vec = sorted(diff_vec,key=lambda x:x[1]) 
        self.ordering = [d[0] for d in diff_vec]
        return 

    def search(self): 
        while not self.fin_stat: 
            next(self) 
        
        for i in range(len(self.best_W)): 
            if self.best_W[i] == 1:
                self.best_W[i] = 10 

    def __next__(self): 
        if self.fin_stat: return 

        q = self.next_element() 
        if type(q) == type(None): 
            self.fin_stat = True 
            return 
        
        V = np.dot(self.samples,q) 
        diff = cumulative_vector_difference(V,True) 

        if self.verbose: 
            print("* V: {}\n* score: {}".format(q,diff /1000)) 

        if diff > self.best_score: 
            self.best_W = deepcopy(q) 
            self.best_score = diff 

    def next_element(self): 
        if self.alter_pattern == "linear": 
            if self.index >=len(self.W): 
                return None
            W_ = deepcopy(self.W) 
            index = self.ordering[self.index] 
            self.W[index] = 10 
            self.index += 1 
            return W_ 
        else: 
            if self.ssi.finished(): 
                return None 
            self.W = next(self.ssi) 
            return self.W 


import numpy as np 
from copy import deepcopy 
from .matrix_methods import is_vector,is_valid_range
from .numerical_generator import prg__single_to_int,prg__LCG,prg_choose_n,prg_decimal,modulo_in_range 
from types import FunctionType,MethodType 

DEFAULT_FIT_TYPE_1_COEFFICIENT_SCALE = [-2/1.2,2/1.2] 
DEFAULT_FIT_TYPE_1_MAX_ADJUSTMENT_STEPS = [1,1000] 

"""
Given a vector V, calculate a vector V2 of length |V| s.t. 
    V * V2 = f. 
"""
class Vector2FloatSolverTypeS1:  

    def __init__(self,V,f,prg,coeff_scale=DEFAULT_FIT_TYPE_1_COEFFICIENT_SCALE,adjustment_steps=None):
        assert is_vector(V)
        assert type(prg) in {FunctionType,MethodType}
        assert is_valid_range(coeff_scale,False,True)
        self.V = V 
        self.f = float(f)
        self.prg = prg 
        self.coeff_scale = coeff_scale
        self.W = np.zeros((len(self.V),)) 
        
        self.adjustment_indices = None 
        self.running_sum = 0
        self.adjustment_steps = adjustment_steps

        self.c = 0
        self.fin_stat = False  

        self.update_log = [] 
        self.initial_weights() 

    def solve(self): 
        while not self.fin_stat: 
            next(self) 


    def __next__(self): 
        if self.fin_stat: 
            return None 

        self.c += 1 
        fit_exact = False 
        if self.c == self.adjustment_steps: 
             fit_exact = True 
             self.fin_stat = True 
        self.fit_next(fit_exact) 

    def output(self): 
        return sum(self.V * self.W) 


    def initial_weights(self): 
        self.W = self.W + 1 
        S = [i for i in range(len(self.V))]
        self.adjustment_indices = self.choose_non_zero_sum_subset(S) 
        self.number_of_adjustment_steps() 
        self.running_sum = sum(self.V * self.W) 

    def subset_coefficient_sum(self,ai): 
        return sum([self.V[i] for i in ai]) 

    def choose_non_zero_sum_subset(self,S,num_attempts=50): 
        if num_attempts < 0: assert False 

        prg_ = prg__single_to_int(self.prg)
        n = modulo_in_range(prg_(),[1,len(S)+1]) 
        q = prg_choose_n(deepcopy(S),n,\
            prg_,is_unique_picker=True)
        stat = round(self.subset_coefficient_sum(q),5) == 0 
        if stat: 
            return self.choose_non_zero_sum_subset(S,num_attempts-1)
        return q 

    def number_of_adjustment_steps(self):
        if type(self.adjustment_steps) == int and self.adjustment_steps > 0: 
            return 
        prg_ = prg__single_to_int(self.prg) 
        self.adjustment_steps = modulo_in_range(prg_(),DEFAULT_FIT_TYPE_1_MAX_ADJUSTMENT_STEPS)
        return

    def fit_next(self,fit_exact:bool): 
        s = 1.0 
        if not fit_exact: 
            s = modulo_in_range(self.prg(),self.coeff_scale)

        diff = self.f - self.output()
        diff = diff * s 

        stat = False 
        while not stat: 
            stat = self.choose_non_zero_sum_subset(self.adjustment_indices)

        q = self.choose_non_zero_sum_subset(self.adjustment_indices)
        cs = self.subset_coefficient_sum(q)

        q2 = diff / cs 

        self.update_vec(q2,q) 
        self.update_log.append((q2,q)) 
        return q2, q 

    def update_vec(self,f,indices): 
        for i in indices: 
            self.W[i] += f  
        return



from .matrix_methods import * 
from .measures import zero_div
from copy import deepcopy
from math import floor 

#------------------------------------------- methods for contiguous representation 
"""
given: a sequence S, either an np.array or a list 
output: sequence S' corresponding to `S`. Every 
    element of S' is a pair of the form 
    [0] value v of S at the j'th index 
    [1] number of contiguous indices, from the 
        j'th index in increasing order, with 
        values that equal v. 

EX: 
S = <0,0,1,2,1,2,3,3,4,4,4>
S' = <(0,2),(1,1),(2,1),(1,1),(2,1),(3,2),(4,3)> 
"""
def contiguous_repr__sequence(S):
    assert is_vector(S) or type(S) == list
    assert len(S) > 0

    q = [[S[0],1]]
    ref = q[-1]

    for i in range(1,len(S)):
        if S[i] == ref[0]:
            ref[1] += 1
        else:
            ref_ = [S[i],1]
            q.append(ref_) 
            ref = ref_
    return q

def indexed_contiguous_repr__sequence(S):
    q = contiguous_repr__sequence(S) 
    i = 0 
    q_ = [] 
    while len(q) > 0: 
        q2 = q.pop(0) 
        q2_ = (q2[0],q2[1],i) 
        i = i + q2[1] 

        q_.append(q2_) 
    return q_ 

def contiguous_repr_size_measure(S,measure=np.var): 
    q = [s[1] for s in S] 
    return measure(q)

def contiguous_repr_size_ratio(S): 
    if len(S) == 0: return 0.0 

    qs = sum([s[1] for s in S])
    q = contiguous_repr_size_measure(S,np.mean) 
    return zero_div(q,qs,0)  


def repeat_cycle_for_length(c,l,ci=0):
    q = np.zeros((l,))

    # go forward 
    for i in range(ci,l,len(c)): 
        x = i 
        x2 = min([i+len(c),l])
        q[x:x2] = c[:x2-x]  

    # go backward
    cx = len(c) - 1 
    for i in range(ci-1,-1,-1):
        q[i] = c[cx] 
        cx = (cx - 1) % len(c) 
    
    return q 

def contiguous_cyclical_difference(V,sv,diff_type="bool"):
    assert diff_type in {"abs","bool"}
    assert len(V) >= len(sv)

    # find the first occurrence of sv in V 
    ir = index_range_of_subvec(V,sv,is_contiguous=True)
    
    if type(ir) == type(None): 
        return float('inf')
    
    vc = repeat_cycle_for_length(sv,len(V),ir[0])
    
    l = np.where(V != vc)[0]

    if diff_type == "bool": return len(l) 
    return np.sum(np.abs(V[l]))

#------------------------------------------ methods for processing (value,freq) vectors 

"""
V is a vector of (value,frequency) pairs 
assumed to be sorted in ascending or descending 
order. Function finds the sequence of all 
(value,frequency) pairs that tie for a place 
in the ranking of frequency. 
"""
def valuefreq_pair_vector__nth_place(V,i=0):

    j = None

    ip = 0
    ref = V[0]
    seq = []
    for i_ in range(1,len(V)):
        if ip == i:
            if ref[1] != V[i_][1]:
                break
            else: 
                seq.append(V[i_][0]) 
            continue
    
        if ref[1] != V[i_][1]:
            ip += 1 
            ref = V[i_] 
    return seq

def valuefreq_pair_vector_to_tie_partition(V):

    j = None
    ip = 0
    ref = V[0]
    seqs = []
    seq = []
    for i_ in range(1,len(V)):
        if ref[1] != V[i_][1]:
            ip += 1 
            ref = V[i_] 
            seqs.append(seq) 
            seq = [V[i_][0]]
        else: 
            seq.append(V[i_][0]) 
    return seqs 

#--------------------------------------------------------------------

"""
most common contiguous subsequence search 
"""
class MCSSearch:

    def __init__(self,L,cast_type=int,is_bfs:bool=True):  
        assert type(L) == list or is_vector(L) 
        assert type(is_bfs) == bool
        self.l = np.array(L)  
        self.cast_type = cast_type 
        self.is_bfs = is_bfs 

        self.preproc()
        # stringized subsequence of L --> index list of occurrence 
        self.subseq_occurrences = defaultdict(list) 
        self.key_queue = [] 
        self.key_cache = []
        return
    
    def preproc(self):
        self.d2index = defaultdict(list) 
        for (i,l_) in enumerate(self.l):
            self.d2index[l_].append(i) 
    
    def most_frequent_(self):
        q = sorted([(k,len(v)) for k,v in self.d2index.items()],key=lambda x:x[1],reverse=True) 

        q_ = q.pop(0)   
        s = set() 
        s |= {q_[0]} 
        while len(q) > 0:
            q2_ = q.pop(0) 
            if q2_[1] != q_[1]:
                break 
            s |= {q2_[0]} 
        return s 
    
    """
    post-search main method 
    """
    def mcs(self):
        x = [(k,len(v)) for k,v in self.subseq_occurrences.items()] 
        x = sorted(x,key=lambda x:x[1])
        q = x.pop(-1)
        s = [string_to_vector(q[0],castFunc=self.cast_type)]
        while len(x) > 0:
            q_ = x.pop(-1)
            if q_[1] == q[1]:
                s_ = string_to_vector(q_[0],castFunc=self.cast_type)
                s.append(s_)
            else: 
                break 
        return s 

    def mcs_nth(self,i=0): 
        x = [(k,len(v)) for k,v in self.subseq_occurrences.items()] 
        x = sorted(x,key=lambda x_:x_[1],reverse=True)
        return valuefreq_pair_vector__nth_place(x,i)
    
    def init_search(self):

        self.subseq_occurrences.clear() 
        self.key_queue.clear()
        self.key_cache.clear() 

        #ss = sorted(self.most_frequent())
        ss = sorted(set(np.unique(self.l)))

        for ss_ in ss: 
            v = self.d2index[ss_] 
            ssv = vector_to_string([ss_],castFunc=self.cast_type)
            self.subseq_occurrences[ssv] = v
            self.key_queue.append(ssv)  
        return
    
    def __next__(self):
        if len(self.key_queue) == 0: 
            return False 
        
        x = self.key_queue.pop(0)
        q = self.extend_subseq(x) 

        self.key_cache.append(x) 

        # sort q by frequency
        q_ = []
        for q2 in q:
            v = self.subseq_occurrences[q2] 
            q_.append((q2,v)) 
        q_ = sorted(q_,key=lambda x:x[1],reverse=True)  
        q_ = [q2[0] for q2 in q_] 

        if self.is_bfs:
            self.key_queue.extend(q_)
        else:
            while len(q_) > 0:
                self.key_queue.insert(0,q_.pop(-1)) 
        return True 

    def extend_subseq(self,subseq_str): 
        q = self.subseq_occurrences[subseq_str] 
        subseq_base = string_to_vector(subseq_str,castFunc=self.cast_type)
        new_subseq_ = set() 
        for q_ in q:
            # get the next index
            i = len(subseq_base) + q_ 

            if i >= len(self.l): continue

            subseq = deepcopy(subseq_base) 
            subseq = np.append(subseq,self.l[i])

            subseq_str_ = vector_to_string(subseq,castFunc=self.cast_type) 
            self.subseq_occurrences[subseq_str_].append(q_)  
            new_subseq_ |= {subseq_str_}
        return new_subseq_ 
    
    """
    pre-search main method 
    """
    def search(self):
        self.init_search() 

        stat = True 
        while stat: 
            stat = self.__next__()
        return 
    
    def kcomplexity(self,keys=None,diff_type="bool",diff_type2="contiguous"):
        assert diff_type in {"bool","abs"}
        assert diff_type2 in {"contiguous","best"}

        #dx = float('inf')
        #rep_set = set()  
        rep_vec = [] 
        if type(keys) == type(None): 
            keys = list(self.subseq_occurrences.keys()) 

        def diff_func(k): 
            v = string_to_vector(k,castFunc=self.cast_type) 
            if diff_type2 == "contiguous": 
                d = contiguous_cyclical_difference(self.l,v,diff_type=diff_type)
                return d 
            ql = len(self.subseq_occurrences[k])
            return len(self.l) - len(v) * ql   
            
        for k in keys:
            d = diff_func(k) 
            rep_vec.append((k,d))
        rep_vec = sorted(rep_vec,key=lambda x:x[1])
        return rep_vec 
    
    def kcomplexity_at_nth_set(self,i,diff_type="bool",diff_type2="contiguous"):
        assert diff_type in {"bool","abs"}
        assert diff_type2 in {"contiguous","best"}

        qs = self.mcs_nth(i) 
        return self.kcomplexity(keys=qs,diff_type=diff_type,diff_type2=diff_type2)

    def default_kcomplexity(self,diff_type="bool",diff_type2="contiguous",\
        basis="most frequent"):

        assert diff_type in {"bool","abs"}
        assert diff_type2 in {"contiguous","best"}
        assert basis in {"most frequent","median"} 

        if basis == "most frequent": 
            qx = self.kcomplexity_at_nth_set(0,diff_type,diff_type2) 
            qx = [qx_[1] for qx_ in qx] 
            return np.mean(qx)

        x = [(k,len(v)) for k,v in self.subseq_occurrences.items()] 
        x = sorted(x,key=lambda x:x[1])

        stat = len(x) % 2
        q = [] 
        # odd
        if stat:
            q_ = x[int(len(x) / 2)] 
            q.append(q_[0]) 
        # even
        else: 
            q_ = x[int(len(x) / 2) - 1] 
            q1_ = x[int(len(x) / 2)] 
            q = [q_[0],q1_[0]] 

        res = self.kcomplexity(keys=q,diff_type=diff_type)
        q = [res_[1] for res_ in res] 
        return np.mean(q) 

def MCS_kcomplexity(L,cast_type,diff_type="bool",\
    diff_type2="contiguous",basis="most frequent"):
    mcs = MCSSearch(L,cast_type=cast_type,is_bfs=True)
    mcs.search() 
    return mcs.default_kcomplexity(diff_type,diff_type2,\
        basis) 




from morebs2.deline import *

"""
Node container for <DLineate22> instance. Used to form trees, 
and these trees are used by class<DLineateMC> for classification. 
"""
class D22Node: 

    def __init__(self,d22,child_nodes):  
        assert type(d22) == DLineate22
        assert type(child_nodes) == list 
        q = []
        for c in child_nodes: 
            assert type(c) == DLineate22
            assert c.parent_idn() == d22.idn 
            q.append(c.idn)
        assert d22.children_idn() == q

        self.d22 = d22
        self.child_nodes = child_nodes

    def idn(self): 
        return self.d22.idn 

    def parent_idn(self): 
        return self.d22.parent_idn() 

    def children_idn(self): 
        return self.d22.children_idn() 

    def add_child_node(self,c): 
        assert type(c) == D22Node 
        assert c.parent_idn() == self.idn() 
        self.child_nodes.append(c) 

    def classify_point(self,x): 
        return self.d22.classify_point(x) 
    
    def classify(self,x): 

        c = self.classify_point(x)
        if c == -1: 
            return -1 

        q = [c] 
        nodes = [self] 
        while len(nodes) > 0: 
            node = nodes.pop(0) 
            cs = [] 

            for cn in node.child_nodes: 
                c2 = cn.classify_point(x) 
                if c2 != -1: 
                    cs.append(c2) 
                    nodes.append(cn)
            q.extend(cs) 

        return q

    def to_node_sequence(self): 
        nodes = [self]
        fin_nodes = []
        while len(nodes) > 0: 
            node = nodes.pop(0)
            nodes.extend(node.child_nodes)         
            fin_nodes.append(node) 

    def node_count(self): 
        '''
        c = 1 
        nodes = [self]

        while len(nodes) > 0: 
            node = nodes.pop(0)
            x = len(node.child_nodes) 
            c += x 
            nodes.extend(node.child_nodes) 
        return c 
        ''' 
        return len(self.to_node_sequence())
    
    def curve_size(self): 
        q = self.to_node_sequence() 
        c = 0 
        for q_ in q: 
            c += len(q_.d22)
        return c 

# NOTE: redundant variable 
"""
V := list 
"""
def most_frequent_label(V):
    d = Counter() 
    for v_ in V: 
        d[v_] += 1 
    d = [(k,v) for k,v in d.items()] 
    return max(d,key=lambda x: x[1])[0] 


"""
A multi-classifier for 2-dimensional dataset built on top of class<DLineate22>. 
"""
class DLineateMC:
    '''
    NOTE: inefficiencies exist. Runtime "slow" for > 200 samples. 
    '''

    def __init__(self,xyl,dmethod="nodup",prg=prg__LCG(-1615.4,-3454.66,-919.16,-9191.4545)):
        self.xyl = xyl
        self.dmethod = dmethod
        self.prg = prg 
        self.xyl_unproc = deepcopy(self.xyl)
        self.d = None
        self.dds = []
        self.didn = "0"

        self.done_cache = [] 
        self.node_cache = [] 
        self.child_cache = [] 
        self.NOT_cache = [] 

        self.fin_stat = False 
        self.cl = None 

    #------------------------------------------------------------------------------------------

    def init_root(self,clockwise=True,max_points_per_edge=100):
        '''
        '''

        self.d = DLineate22(deepcopy(self.xyl_unproc),clockwise=clockwise,dmethod = self.dmethod,\
            idn=self.didn,target_min_label=False,max_points_per_edge=max_points_per_edge,\
            prg=self.prg)

        self.increment_didn()
        _,_,_,met = self.d.full_process(None)
        self.node_cache.append((self.d,met)) 
        return

    def __next__(self):
        if self.fin_stat: 
            return False 

        if len(self.node_cache) == 0: 
            self.fin_stat = True 
            return False 

        parent_d22,parent_met = self.node_cache.pop(0) 
        self.process_D22(parent_d22,parent_met) 
        return True 

    def process_D22(self,parent_d22,parent_met,clockwise=True):
        D_fpos,met_fpos = self.process_D22_(parent_d22,parent_met,clockwise,process_type="fpos")
        
        if type(D_fpos) != type(None): 
            D_fpos,met_fpos = self.post_process_d22(parent_d22,D_fpos,met_fpos)
            self.node_cache.append((D_fpos,met_fpos)) 

        self.process_D22_remaining_NOTs(parent_d22,parent_met,clockwise)
        self.done_cache.append(parent_d22) 
        return

    def process_D22_(self,parent_d22,parent_met,clockwise=True,process_type="fpos",not_indices=None): 
        assert process_type in {'fpos','not'}
        # process first child 

        if process_type == "fpos": 
            xyl = parent_d22.xyl[parent_met.fpos_indices] 
        else: 
            if type(not_indices) == type(None): 
                not_indices = parent_met.NOT_indices() 
            
            xyl = parent_d22.xyl[not_indices] 

        if len(xyl) == 0: 
            return None,None 

        D = DLineate22(xyl,clockwise=clockwise,dmethod = self.dmethod,\
            idn=self.didn,target_min_label=False,max_points_per_edge=100,\
            prg=self.prg)
        self.increment_didn() 

        idn = None if process_type == "not" else parent_d22.idn
        _,_,_,met2 = D.full_process(idn,xyl) 
        return D,met2

    def process_D22_remaining_NOTs(self,parent_d22,parent_met,clockwise=True): 

        met_seq = [parent_met]
        d22_seq = [parent_d22] 

        not_indices = parent_met.NOT_indices()

        def fetch_not_contained_indices(ref_met): 
            NOT = set(ref_met.NOT_indices())
            for x in met_seq: 
                not2 = set(x.NOT_indices())
                NOT = NOT.intersection(not2) 
            return sorted(NOT) 

        while len(not_indices) > 0: 
            D,met2 = self.process_D22_(parent_d22,parent_met,clockwise,"not",not_indices)   
            if type(D) == type(None): continue 

            not_indices = fetch_not_contained_indices(met2) 
            d22_seq.append(D)
            met_seq.append(met2) 

        met_seq.pop(0) 
        d22_seq.pop(0) 

        while len(d22_seq) > 0: 
            d,m = d22_seq.pop(0),met_seq.pop(0) 
            d,m = self.post_process_d22(parent_d22,d,m)
            self.node_cache.append((d,m))
        return

    # NOTE: function does nothing
    def post_process_d22(self,parent_d22,d22,d22_metric): 
        """
        print("PP ",parent_d22.idn,d22.idn) 
        print("DXDX")
        print(parent_d22.xyl)
        print("dxdx")
        print(d22.xyl)
        print("D22 METRIC")
        print(d22_metric)
        data_IN = parent_d22.xyl[d22_metric.contained_indices]
        _,_,_,d22_metric = d22.analyze_delineation(data_IN) 
        d22.xyl = data_IN 

        print("DATA IS")
        print(data_IN) 
        """
        return d22,d22_metric 

    def increment_didn(self):
        self.didn = str(int(self.didn) + 1)

    def decrement_didn(self):
        self.didn = str(int(self.didn) - 1)

    def to_graph(self):
        x = {}
        for q in self.done_cache: 
            x[q.idn] = D22Node(q,[])
        
        S = [] 

        for q in self.done_cache: 
            p = q.parent_idn()
            if type(p) == type(None): 
                S.append(x[q.idn]) 
                continue 
            x[p].add_child_node(x[q.idn])
        self.cl = S 
        return

    """
    main method #4

    outputs a classification map, 
        DLineate22 idn -> {label sequence}
    """
    def classify_(self,x): 
        cmap = {} 
        for q in self.cl: 
            c = q.classify(x)
            cmap[q.idn()] = c 
        return cmap 

    """
    main method #3 

    outputs a sequence of most frequent label 
    per <DLineate22>
    """
    def classify(self,x): 
        cmap = self.classify_(x) 
        C = [] 
        for k,v in cmap.items(): 
            if v == -1: 
                continue
            q = v[-1]
            C.append(q)  

        if len(C) == 0: return -1 
        return sorted(C) 

    """
    main method #2 
    
    classifies vector to 1 label, most frequent in 
    vector from method<classify>. 
    """
    def one_classification(self,x): 
        v = self.classify(x) 
        if v == -1: return -1 
        return most_frequent_label(v)

    """
    main method #1: calculate classifier 
    """
    def full_run(self): 
        self.init_root() 
        stat = True 
        while stat: 
            stat = next(self) 
        self.to_graph() 

    def node_count(self): 
        cmap = {} 
        for q in self.cl: 
            cmap[q.idn()] = q.node_count() 
        return cmap 

DEFAULT_D22MC_MAX_POINTS_PER_LABEL = 20 

"""
An approximator built on top of class<DLineateMC>. By default, takes at most 20 
points per label of two-dimensional dataset. 
"""
class DLineateMCApprox(DLineateMC): 

    def __init__(self,xyl,dmethod="nodup",prg=prg__LCG(-1615.4,-3454.66,-919.16,-9191.4545)):
        self.D = xyl 
        xyl = approximate_points_for_delineation(self.D,\
            max_points_per_label=DEFAULT_D22MC_MAX_POINTS_PER_LABEL,prg=prg) 
        super().__init__(xyl,dmethod,prg)        
        return 

"""
Uses m <= `max_approximators` approximators to find a multi-class classifier. 
Built on top of class<DLineateMCApprox>. 
"""
class DLineateMCApproximators: 

    def __init__(self,xyl,max_approximators,dmethod="nodup",\
        prg=prg__LCG(-1615.4,-3454.66,-919.16,-9191.4545)):
        assert type(max_approximators) == int and max_approximators > 0 

        self.xyl = xyl 
        self.rem_xyl = deepcopy(xyl)
        self.max_approximators = max_approximators
        self.dmethod = dmethod 
        self.prg = prg 
        self.approximators = [] 
        self.fin_stat = False 

    """
    main method #1
    """
    def full_run(self): 
        while not self.fin_stat: 
            next(self) 

    def __next__(self): 
        if self.fin_stat: 
            return False 

        if len(self.approximators) == self.max_approximators: 
            self.fin_stat = True 
            return False 
        
        self.one_approximator()
        self.gauge_approximator() 

    def one_approximator(self): 
        dma = DLineateMCApprox(self.rem_xyl,dmethod=self.dmethod,prg=self.prg)
        dma.full_run() 
        self.approximators.append(dma) 

    def gauge_approximator(self): 
        dma = self.approximators[-1] 
        indices = [] 
        for i,x in enumerate(self.rem_xyl): 
            l_ = dma.one_classification(x[:2])
            if l_ == -1: 
                indices.append(i)
        if len(indices) == 0: 
            self.fin_stat = True 
            return 

        self.rem_xyl = self.rem_xyl[indices]

    """
    main method #2 
    """
    def one_classification(self,x): 
        l = [] 
        for a in self.approximators: 
            l_ = a.one_classification(x)
            if l_ == -1: continue 
            l.append(l_) 

        if len(l) == 0: return -1 
        return most_frequent_label(l) 



from .graph_basics import * 

"""
Auxiliary method for <calculate_dependency_order_for_IsoRing_list>.
"""
def order_element_in_dependency_order(depANDcodep_map,ooc,index):

    cds = ooc[index]
    # get all dependencies 
    dep = set() 
    for idn in cds:  
        dep |= depANDcodep_map[idn][0] 

    # case: codependencies cannot be dependencies 
    inter = dep.intersection(ooc[index]) 
    if len(inter) > 0: 
        return None,False  

    #print("DEP FOR {}: ".format(ooc[index],dep))  
    qi = index 
    x = None  
    for i in range(index + 1,len(ooc)): 
        inter = dep.intersection(ooc[i])  
        #print("OOC: ",ooc[i]) 
        if len(inter) > 0: 
            #print("YES") 
            x = i 
    
    if type(x) == type(None): 
        return -1,True 

    # pop element 
    q = ooc.pop(index) 
    ooc.insert(x,q) 
    return x,True   

def calculate_codependency_order(depANDcodep_map): 
    ooc = [] 
    num_no_deps = 0 
    # start with codependencies and verify
    elements = sorted(depANDcodep_map.keys()) 
    for elem in elements:
        
        ds = depANDcodep_map[elem][0] 
        if len(ds) == 0: 
            num_no_deps += 1 

        cds = depANDcodep_map[elem][1]
        cds = cds | {elem}  
        qi = index_of_element_in_setseq(ooc,elem)  

        if type(qi) == type(None): 
            ooc.append(cds) 
            continue 

        if ooc[qi] != cds: 
            return None,False,num_no_deps

    return ooc,True,num_no_deps

def calculate_dependency_order_(ooc,depANDcodep_map):  
    qx = deepcopy(ooc) 
    for i in range(len(ooc)):
        rx = qx[i]
        rx_ = next(iter(rx)) 
        j = index_of_element_in_setseq(ooc,rx_)
        index,stat = order_element_in_dependency_order(depANDcodep_map,ooc,j) 

        # case: error, co-dep cannot be dep 
        if not stat: 
            return None,False 
    return ooc,True 

"""
Calculates a dependency ordering using depANDcodep_map,
    element idn -> (dependency set, codependency set). 
During this calculation, determines if dependencies and 
codependencies of every <IsoRing> result in consistentn 
<IsoRingedChain>. 
"""
def calculate_dependency_order(depANDcodep_map):  

    ooc,stat,num_no_deps = calculate_codependency_order(depANDcodep_map) 
    if not stat: 
        return ooc,stat 

    # order elements in first scan
    for _ in range(num_no_deps): 
        ooc,stat = calculate_dependency_order_(ooc,depANDcodep_map)
        if not stat: 
            return ooc,stat 

    # check for contradictions in second scan
    for i in range(len(ooc)): 
        index, stat = order_element_in_dependency_order(depANDcodep_map,ooc,i)
        if index != -1: 
            return None,False 

    return ooc,True 

class DependencyGraph:

    def __init__(self,depANDcodep_map): 
        self.depANDcodep_map = depANDcodep_map
        self.setseq = None
        self.G = None 
        self.heads = None  
        self.ooc,self.stat,self.num_no_deps = None,None,None 
        self.active_proc_finstat = False 
        return 

    def process_order(self,require_proper_order:bool=True): 
        self.active_proc_finstat = True  
        if require_proper_order: 
            self.ooc,self.stat = calculate_dependency_order(self.depANDcodep_map)
            self.active_proc_finstat = False  
            return 

        self.ooc,self.stat,self.num_no_deps = calculate_codependency_order(self.depANDcodep_map)
        return

    def next_order(self): 
        if not self.active_proc_finstat: return False 

        ooc,self.stat = calculate_dependency_order_(deepcopy(self.ooc),self.depANDcodep_map)
        if not self.stat: 
            self.active_proc_finstat = False 

        if ooc == self.ooc: 
            self.active_proc_finstat = False 
        elif type(ooc) == type(None): 
            self.active_proc_finstat = False 
        else: 
            self.ooc = ooc 

    def to_graph(self): 
        assert type(self.ooc) != type(None) 

        self.G = defaultdict(set) 
        for i in range(len(self.ooc)): 
            self.next_nodeset(i)
        return self.G 

    def next_nodeset(self,index): 
        if index == 0: self.heads = deepcopy(self.ooc[index]) 

        for x in self.ooc[index]: 
            self.G[x] = self.ooc[index] - {x} 

        if index == 0: 
            return 

        q = self.ooc[index-1] 
        for x in self.ooc[index]: 
            for x2 in q: 
                self.G[x2] |= {x} 
        return


"""
file for simulation of recursive processes
"""
from types import MethodType,FunctionType


"""
Simulates a recursive method. 

Is comprised of n conditional statements, n > 0, and a possible NOT conditional. 
Every conditional has 4 elements: 
[0] boolean function : accept or reject parameters.
    - null for NOT conditional.  
[1] parameter delta function: modifies parameters. 
    - optional; can be null. 
[2] output function: maps parameters to an output value. 
    - optional; can be null. 
[3] termination function: outputs boolean for termination of recursive process on `parameters.

The `pre_delta_function` P is optional. 

----------------------------------------------------------------------------------------------

Process is first set with starting `parameters`, via method<load_parameters>. 

For every recursive call (method<__next__>): 
- If P is not null, apply 
    `parameters` = P(`parameters`). 
- Iterate through the conditionals c_i in C: 
    if c_i[0](`parameters`): 
        Process `parameters` using that conditional. 
        Quit call.
- If no conditionals apply, then apply the NOT conditional. 
"""
class SimulatedRecursionNode:

    def __init__(self,pre_delta_function,conditionals,not_conditional): 
        assert type(pre_delta_function) in {type(None),MethodType,FunctionType}
        self.pre_delta_function = pre_delta_function

        assert len(conditionals) > 0 
        for x in conditionals: 
            self.assert_proper_cond(x,False) 
        self.conditionals = conditionals

        self.assert_proper_cond(not_conditional,True)
        self.not_conditional = not_conditional
        self.parameters = None
        self.output_seq = [] 

        self.terminated = False 
        self.c = 0 
        return

    def set_pre_delta_function(self,p): 
        assert type(p) in {type(None),MethodType,FunctionType}
        self.pre_delta_function = p 

    def add_conditional(self,c): 
        self.assert_proper_cond(c,False) 
        self.conditionals.append(c) 
        return 

    def replace_conditional(self,c,index): 
        assert 0 <= index < len(self.conditionals)
        self.assert_proper_cond(c,False) 
        self.conditionals[index] = c         
        return

    def set_NOT_conditional(self,c): 
        self.assert_proper_cond(c,True) 
        return

    def assert_proper_cond(self,x,is_not_cond:bool):
        if type(x) == type(None) and is_not_cond: return 

        assert type(x) == tuple  
        assert len(x) == 4 
        
        # boolean condition 
        if not is_not_cond:
            assert type(x[0]) in {MethodType,FunctionType}
        else: 
            assert type(x[0]) == type(None) 

        # parameter delta 
        assert type(x[1]) in {MethodType,FunctionType,type(None)} 
        # output function 
        assert type(x[2]) in {MethodType,FunctionType,type(None)}
        # terminating condition 
        assert type(x[3]) in {MethodType,FunctionType} 
        return

    def load_parameters(self,parameter_seq): 
        assert type(parameter_seq) == list 
        self.parameters = parameter_seq 
        self.output_seq.clear() 
        self.terminated = False
        self.c = 0  
        return

    def process_conditional(self,i): 
        x = self.conditionals[i]
        return self.process_conditional_(x)

    def process_conditional_(self,x): 
        if type(x) == type(None): return False 

        q = x[0] 

        # NOT conditional 
        if type(q) == type(None): 
            self.one_proc(x) 
            return True 
        
        if q(*self.parameters): 
            self.one_proc(x) 
            return True 

        return False 

    def one_proc(self,x): 
        q1 = x[1] 
        if type(q1) != type(None): 
            self.parameters = q1(*self.parameters) 

        q2 = x[2]  
        if type(q2) != type(None): 
            output = q2(*self.parameters) 
            self.output_seq.append(output) 
        
        q3 = x[3] 
        self.terminated = q3(*self.parameters)  
        return 

    def __next__(self): 
        if self.terminated: return 
        self.c += 1

        if type(self.pre_delta_function) != type(None): 
            self.parameters = self.pre_delta_function(*self.parameters)

        stat_ = False 
        for i in range(len(self.conditionals)): 
            stat_ = self.process_conditional(i)
            if stat_: 
                break 

        if not stat_: 
            stat_ = self.process_conditional_(self.not_conditional)
        return stat_ 


from .classification_test_cases import * 
from morebs2.modular_labeller import * 
from morebs2.odc_extended import * 
import unittest

'''
py -m morebs2.tests.odc_extended_test  
'''
class AdditiveAdjustedRecursiveODCClass(unittest.TestCase):

    def test__AdditiveAdjustedRecursiveODC__score_accuracy_case_1(self):

        D = OneDimClassifier_test_dataset_1() 
        L = np.array([label_vector__type_binary_alt_sum(v) for v in D]) 

        # subcase 1
        prg = prg__LCG(-56.7,122.3,-31.6,-3121.66) 
        rodc = AdditiveAdjustedRecursiveODC(D,L,prg,0)
        rodc.fit() 

        c = rodc.score_accuracy(D,L)
        c2 = rodc.score_accuracy(D,L)
        c3 = rodc.score_accuracy(D,L)
        assert c == c2 == c3 == 1000, "got {},{}".format(c,c2)

        # subcase 2 
        rodc = AdditiveAdjustedRecursiveODC(D,L,prg,1)
        rodc.fit() 

        c = rodc.score_accuracy(D,L)
        c2 = rodc.score_accuracy(D,L)
        assert c == c2 == 904, "got {}".format(c)
        return

    """
    tests <ODCAdditiveAdjustmentNode> with half of its contiguous information 
    missing. 

    Diminished accuracy from 1000 to 559. 
    """ 
    def test__AdditiveAdjustedRecursiveODC__score_accuracy_case_2(self):
        D = OneDimClassifier_test_dataset_1() 
        L = np.array([label_vector__type_binary_alt_sum(v) for v in D]) 

        # subcase 1
        prg = prg__LCG(-56.7,122.3,-31.6,-3121.66) 
        rodc = AdditiveAdjustedRecursiveODC(D,L,prg,0)
        rodc.fit() 

        # try removing some of the indices
        node = rodc.root_ 
        cs = node.contiguous_sequence
        cs_ = [cs.pop(0),cs.pop(-1)]
        qx = prg_choose_n(cs,len(cs) // 2,prg__single_to_int(prg),is_unique_picker=True)
        qx = sorted(qx + cs_,key=lambda x: x[2]) 
        node.contiguous_sequence = qx 

        c = rodc.score_accuracy(D,L)
        assert c == 590,"got {}".format(c)

class MBAAARecursiveODCClass(unittest.TestCase):

    """
    tests demonstrate the varying accuracy of classifiers 
    based on different memory parameters, as well as 
    identifical parameters but different PRNG-based decisions. 
    """
    def test__MBAAARecursiveODC__score_accuracy_case_1(self):
        D = OneDimClassifier_test_dataset_1() 
        L2 = np.array([label_vector__type_binary_alt_sum(v) for v in D]) 
        
        ##q = deepcopy(D) 
        ##L2_ = deepcopy(L2)

        # subcase 1
        prg = prg__LCG(-56.7,122.3,-31.6,-3121.66) 

        rodc2 = MBAAARecursiveODC(D,L2,prg,0,max_mem_sz=2000,mem_ratio_per_node=0.95) 
        rodc2.fit() 
        c2 = rodc2.score_accuracy(D,L2)

        rodc3 = MBAAARecursiveODC(D,L2,prg,0,max_mem_sz=2000,mem_ratio_per_node=0.9) 
        rodc3.fit() 
        c3 = rodc3.score_accuracy(D,L2)

        rodc4 = MBAAARecursiveODC(D,L2,prg,0,max_mem_sz=2000,mem_ratio_per_node=0.75) 
        rodc4.fit() 
        c4 = rodc4.score_accuracy(D,L2)

        rodc5 = MBAAARecursiveODC(D,L2,prg,0,max_mem_sz=2000,mem_ratio_per_node=0.6) 
        rodc5.fit() 
        c5 = rodc5.score_accuracy(D,L2)

        rodc6 = MBAAARecursiveODC(D,L2,prg,0,max_mem_sz=2000,mem_ratio_per_node=0.5) 
        rodc6.fit() 
        c6 = rodc6.score_accuracy(D,L2)

        rodc7 = MBAAARecursiveODC(D,L2,prg,0,max_mem_sz=2000,mem_ratio_per_node=0.5) 
        rodc7.fit() 
        c7 = rodc7.score_accuracy(D,L2)

        rodc8 = MBAAARecursiveODC(D,L2,prg,0,max_mem_sz=2000,mem_ratio_per_node=0.95) 
        rodc8.fit() 
        c8 = rodc8.score_accuracy(D,L2)

        rodc9 = MBAAARecursiveODC(D,L2,prg,1,max_mem_sz=2000,mem_ratio_per_node=0.95) 
        rodc9.fit() 
        c9 = rodc9.score_accuracy(D,L2)

        CX = [c2,c3,c4,c5,c6,c7,c8,c9] 
        assert [987, 622, 473, 431, 501, 401, 675, 963]


    """
    tests demonstrate varying accuracy, based on PRNG decisions, 
    for classifier with `max_mem_sz`=400,mem_ratio_per_node=0.6. 
    """
    def test__MBAAARecursiveODC__score_accuracy_case_2(self):

        D = OneDimClassifier_test_dataset_1() 
        L2 = np.array([label_vector__type_binary_alt_sum(v) for v in D]) 
                
        prg = prg__LCG(56.7,122.3,-31.6,2121.66) 
        rodc = MBAAARecursiveODC(D,L2,prg,1,max_mem_sz=400,mem_ratio_per_node=0.6) 
        rodc.fit() 
        c10 = rodc.score_accuracy(D,L2)

        rodc2 = MBAAARecursiveODC(D,L2,prg,1,max_mem_sz=400,mem_ratio_per_node=0.6) 
        rodc2.fit() 
        c11 = rodc2.score_accuracy(D,L2)

        rodc3 = MBAAARecursiveODC(D,L2,prg,1,max_mem_sz=400,mem_ratio_per_node=0.6) 
        rodc3.fit() 
        c12 = rodc3.score_accuracy(D,L2)

        rodc4 = MBAAARecursiveODC(D,L2,prg,1,max_mem_sz=400,mem_ratio_per_node=0.6) 
        rodc4.fit() 
        c13 = rodc4.score_accuracy(D,L2)

        assert [c10,c11,c12,c13] == [372,312,359,365]

if __name__ == '__main__':
    unittest.main()


from .odc_nodes import * 

"""
A dataset classifier, based on decision tree format. 
Uses the classifier <OneDimClassifier> as a unit node. 
"""
class RecursiveOneDimClassifier: 

    def __init__(self,D,L,prg=None,pscheme=0,verbose:bool=False): 
        assert is_2dmatrix(D) 
        assert is_vector(L) 
        self.D = D 
        self.L = L 
        self.fin_stat = False 

        if type(prg) not in {MethodType,FunctionType}: 
            self.prg = default_std_Python_prng()
        else: 
            self.prg = prg 
        assert pscheme in {0,1} or type(pscheme) in {MethodType,FunctionType}
        self.pscheme = pscheme 

        assert type(verbose) == bool 
        self.verbose = verbose 
        self.root = None 

        # every element is 
        # [0] ODCNode
        # [1] P := dict, label -> {indices of label in sub-data}
        # [2] E := dict, label -> number of positive classification errors in D  
        # [3] D := 2-d matrix, sub-data 
        # [4] L := vector, labels corresponding to sub-data 
        self.node_cache = [] 
        self.current = None 

        self.sample_cls_path = [] 

    def clear_data(self): 
        self.D = None 
        self.L = None 

    #--------------------------- main methods 

    """
    main method #3 
    """
    def score_accuracy(self,D,L): 
        assert len(D) == len(L) 

        c = 0 
        for (x,l) in zip(D,L): 
            l_ = self.classify(x) 
            if l_ == l: 
                c += 1 
        return c 

    """
    main method #2 
    """
    def classify(self,v): 
        q = self.root 
        l = None 
        self.sample_cls_path.clear() 
        while type(q) != type(None): 
            l,q = q.classify(v) 
            self.sample_cls_path.append((l,q))  
        return l 

    #------------------------------------- used to find classification solution 

    """
    main method #1 
    """
    def fit(self): 
        self.init_root() 
        while not self.fin_stat: 
            self.next_process() 
        self.clear_data() 

    def init_root(self): 
        self.one_process(self.D,self.L)  
        self.root = self.node_cache[0][0] 

    def next_process(self): 
        if self.fin_stat: 
            return 

        # case: load a new node of interest 
        if type(self.current) == type(None): 
            if len(self.node_cache) == 0: 
                self.fin_stat = True 
                return 
            self.current = self.node_cache.pop(0) 

        # iterate through and clear out any labels with 0 error s
        keys = sorted(self.current[2].keys())
        key0 = [] 
        for k in keys: 
            if len(self.current[2][k]) == 0:
                del self.current[2][k] 
                key0.append(k) 
        rem_keys = sorted(set(keys) - set(key0)) 

        # case: perfect labeling 
        if len(rem_keys) == 0: 
            if self.verbose: 
                print("\t\t** one node")
                print(self.current[0]) 
            self.current = None 
            return 

        i = prg__single_to_int(self.prg)() % len(rem_keys) 
        subset_label = rem_keys[i] 
        del self.current[2][subset_label]

        D,L = [],[]
        for i in self.current[1][subset_label]: 
            D.append(self.current[3][i]) 
            L.append(self.current[4][i]) 
        D,L = np.array(D),np.array(L)
        odcn2,P2,E2 = self.one_child(self.current[0],subset_label,D,L)

    def one_process(self,D,L,previous_indices=set()): 
        odc,P,E = self.one_classification_process_(D,L,previous_indices)
        if type(odc) == type(None): 
            return None,None,None 

        self.node_cache.append((odc,P,E,D,L)) 
        return odc,P,E 

    """
    return: 
    [0] ODCNode 
    [1] dict, label -> {indices}
    [2] dict, label -> {mislabeled indices} 
    """ 
    def one_classification_process_(self,D,L,previous_indices=set()):  

        odc,index = OneDimClassifier.one_OneDimClassifier(D,L,previous_indices,self.prg,self.pscheme)
        if type(odc) == type(None): 
            return None,None,None 

        # partition by classification 
        P = defaultdict(set) 
        E = defaultdict(set)#int)
        for i,x in enumerate(D): 
            l = odc.classify(x) 
            P[l] |= {i} 
            if l != L[i]: 
                #E[l] += 1
                E[l] |= {i} 
        odcn = ODCNode(odc.clf,dict(),previous_indices | {index}) 
        return odcn,P,E 

    def one_child(self,odcn,subset_label,D,L): 
        previous_indices = odcn.previous_indices
        odcn2,P,E = self.one_process(D,L,previous_indices)
        if type(odcn2) ==type(None): return None,None,None 
        odcn.add_nextnode(odcn2,subset_label) 
        return odcn2,P,E 



from morebs2.sr_struct import * 
import unittest

def SimulatedRecursionNode__sample_T(): 

    def pre_delta_function(x0,x1,x2): 
        x0 += 1 
        x1 -= 1 
        x2 += 1 
        return [x0,x1,x2]

    c0 = lambda x0,x1,x2: x0 + x1 + x2 > 100 
    c1 = None 
    c2 = lambda x0,x1,x2: x0 + x1 + x2 
    c3 = lambda x0,x1,x2: (x0 + x1 + x2) % 131 == 0

    C = (c0,c1,c2,c3)

    c4 = lambda x0,x1,x2: (x0 + x1 + x2) % 78 == 0
    C2 = (None,None,c2,c4)

    S = SimulatedRecursionNode(pre_delta_function,[C],C2) 
    return S 


def SimulatedRecursionNode__sample_U(): 
    S = SimulatedRecursionNode__sample_T() 

    def pre_delta_function(x0,x1,x2): 
        x0 *= 2 
        x1 -= 117 
        x2 += 1919
        return [x0,x1,x2]

    S.set_pre_delta_function(pre_delta_function)

    return S 

#########################################################

'''
py -m morebs2.tests.sr_struct_test   
'''
class TestSimulatedRecursionNode(unittest.TestCase):

    def test__SimulatedRecursionNode__next__case_1(self):#

        S = SimulatedRecursionNode__sample_T()

        p = [23,-5,16] 
        S.load_parameters(p) 

        while not S.terminated: 
            next(S) 
        assert sum(S.parameters) == 78 
        assert S.parameters == [67, -49, 60]
        assert len(S.output_seq) == 44, "got {}".format(len(S.output_seq))
        assert S.output_seq[-1] == 78 

    def test__SimulatedRecursionNode__next__case_2(self):#

        S = SimulatedRecursionNode__sample_U()

        p = [23,-5,16] 
        S.load_parameters(p) 

        while not S.terminated: 
            next(S) 

        q = sum(S.parameters)
        assert q == 444884701618183536292024147
        assert q % 131 == 0 

if __name__ == '__main__':
    unittest.main()



from morebs2.numerical_space_data_generator import * 

def OneDimClassifier_test_dataset_1(): 

    bounds = np.zeros((4,2)) 
    bounds[:,0] -= 10 
    bounds[:,1] += 10 

    startPoint = np.zeros((4,)) - 10 
    columnOrder = np.array([0,1,2,3]) 
    ssi_hop = 3 

    rch2 = sample_rch_1_with_update(deepcopy(bounds),deepcopy(bounds),ssi_hop,0.1)
    cv = 0.4
    rssi2 = ResplattingSearchSpaceIterator(bounds, startPoint, \
        columnOrder, SSIHop = ssi_hop,resplattingMode = ("relevance zoom",rch2), additionalUpdateArgs = (cv,))

    ##xr = np.array([next(rssi) for _ in range(1000)]) 
    xr2 = np.array([next(rssi2) for _ in range(1000)])
    return xr2 


from morebs2.v2f_solver import * 
from morebs2.numerical_generator import prg__single_to_nvec
import unittest

### lone file test 
"""
py -m morebs2.tests.v2f_solver_test
"""
###
class Vector2FloatSolverTypeS1Class(unittest.TestCase):

    def test__Vector2FloatSolverTypeS1__solve__case_1(self):

        V = np.array([5,0,6,7,-5,-13]) 
        prg = prg__LCG(45,-677,899,1900.05)

        f = 1200 
        f2 = -2456 

        # subcase 1 
        vs = Vector2FloatSolverTypeS1(V,f,prg)
        vs.solve() 
        assert round(abs(vs.output() - f),4) == 0 
        print("solved {} --> \n\t{} *\n\t{}".format(f,V,np.round(vs.W,5))) 

        # subcase 2 
        vs2 = Vector2FloatSolverTypeS1(V,f2,prg)
        vs2.solve() 
        assert round(abs(vs2.output() - f2),4) == 0 
        print("solved {} --> \n\t{} *\n\t{}".format(f2,V,np.round(vs2.W,5))) 

        # subcases 3-6
        q = [5,4,10,8] 
        for q_ in q: 
            V2 = prg__single_to_nvec(prg,q_)()
            f_ = (prg() - prg()) * (prg() - prg())
            vs3 = Vector2FloatSolverTypeS1(V2,f_,prg)
            vs3.solve() 
            assert round(abs(vs3.output() - f_),4) == 0 
            print("solved {} --> \n\t{} *\n\t{}".format(f_,V2,np.round(vs3.W,5)))  

    def test__Vector2FloatSolverTypeS1__solve__case_2(self):

        V = np.array([5,0,0,7,0,0])  
        prg = prg__LCG(45,-677,899,1900.05)

        f = 44567 
        f2 = -845611 

        vs = Vector2FloatSolverTypeS1(V,f,prg)
        vs.solve() 
        assert round(abs(vs.output() - f),4) == 0 

        vs2 = Vector2FloatSolverTypeS1(V,f2,prg)
        vs2.solve() 
        assert round(abs(vs2.output() - f2),4) == 0 

        q = [5,4,10,8] 
        for q_ in q: 
            V2 = prg__single_to_nvec(prg,q_)()
            f_ = (prg() - prg()) * (prg() - prg())
            vs3 = Vector2FloatSolverTypeS1(V2,f_,prg)
            vs3.solve() 
            assert round(abs(vs3.output() - f_),4) == 0 


if __name__ == '__main__':
    unittest.main()


from morebs2 import deline
from collections import Counter 
import numpy as np
import unittest

'''
py -m morebs2.tests.deline_test  
'''
class DelineClass(unittest.TestCase):

    def test__DLineate22__collect_break_points__case_1(self):
        print("DLineat22 test 1")
        data = deline.test_dataset__Dlineate22_1()
        dl = deline.DLineate22(data)
        dl.preprocess()
        dl.collect_break_points()

        l = np.array([[ 5.,  0.],\
        [5.,6.],\
        [ 5.,  9.],\
        [ 5., 12.],\
        [ 5., 15.]])
        
        r = np.array([[20., 15.],\
        [15., 12.],\
        [15.,  9.],\
        [25, 7.5],\
        [20.,  0.]])

        t = np.array([[ 5. , 15.],\
        [15. , 12.],\
        [20. , 15.],\
        [25. ,  7.5]])

        b = np.array([[25. ,  7.5],\
        [20. ,  0.],\
        [15. ,  9.],\
        [ 5. ,  0.]])

        assert np.all(dl.d.d['l'] == l),"got {}".format(dl.d.d['l'])
        assert np.all(dl.d.d['r'] == r),"got {}".format(dl.d.d['r'])
        assert np.all(dl.d.d['t'] == t),"got {}".format(dl.d.d['t'])
        assert np.all(dl.d.d['b'] == b),"got {}".format(dl.d.d['b'])

    def test__DLineate22__collect_break_points__case_2(self):

        print("DLineat22 test 2")

        # clockwise 
        data = deline.test_dataset__Dlineate22_2()
        dl = deline.DLineate22(data)
        dl.preprocess()
        dl.collect_break_points()

        l = np.array([[ 5.,  0.],\
            [ 5., 15.]])

        r = np.array([[20., 15.],\
            [20.,  0.]])

        t = np.array([[ 5., 15.],\
            [20., 15.]])

        b = np.array([[20.,  0.],\
            [ 5.,  0.]])

        assert np.all(dl.d.d['l'] == l),"got {}".format(dl.d.d['l'])
        assert np.all(dl.d.d['r'] == r),"got {}".format(dl.d.d['r'])
        assert np.all(dl.d.d['t'] == t),"got {}".format(dl.d.d['t'])
        assert np.all(dl.d.d['b'] == b),"got {}".format(dl.d.d['b'])

        # counter-clockwise 
        data = deline.test_dataset__Dlineate22_2()
        dl = deline.DLineate22(data,False)
        dl.preprocess()
        dl.collect_break_points()

        l = np.array([[ 5., 15.],\
            [ 5.,  0.]])

        r = np.array([[20.,  0.],\
            [20., 15.]])

        t = np.array([[20., 15.],\
            [ 5., 15.]])

        b = np.array([[ 5.,  0.],\
            [20.,  0.]])

        assert np.all(dl.d.d['l'] == l),"got {}".format(dl.d.d['l'])
        assert np.all(dl.d.d['r'] == r),"got {}".format(dl.d.d['r'])
        assert np.all(dl.d.d['t'] == t),"got {}".format(dl.d.d['t'])
        assert np.all(dl.d.d['b'] == b),"got {}".format(dl.d.d['b'])

    def test__DLineate22__classify_points__case_3(self):
        print("DLineat22 test 3")

        td3 = deline.test_dataset__Dlineate22_3()
        p = [12.5,7.5]

        dl = deline.DLineate22(np.copy(td3),dmethod="nocross")
        dl.preprocess()
        dl.collect_break_points()

        c = dl.d.classify_point(p)    
        assert c == 0.0, "dmethod=nocross incorrectly classifies"

        dl = deline.DLineate22(np.copy(td3),dmethod="nodup")
        dl.preprocess()
        dl.collect_break_points()

        c = dl.d.classify_point(p)    
        assert c == 0.0, "dmethod=nodup incorrectly classifies"

        dl = deline.DLineate22(np.copy(td3),dmethod="nojag")
        dl.preprocess()
        dl.collect_break_points()

        c = dl.d.classify_point(p)    
        assert c == 0.0, "dmethod=nojag incorrectly classifies"

    def test__DLineate22__classify_points__case_4(self):

        print("DLineat22 test 4")

        td4 = deline.test_dataset__Dlineate22_4()
        dl = deline.DLineate22(np.copy(td4),dmethod="nodup")
        dl.preprocess()
        dl.collect_break_points()

        p = [12.5,7.5]
        c = dl.d.classify_point(p)
        assert c == 0, "misclassification for case 4, dmethod=nodup"

        dl = deline.DLineate22(np.copy(td4),dmethod="nocross")
        dl.preprocess()
        dl.collect_break_points()

        p = [12.5,7.5]
        c = dl.d.classify_point(p)
        assert c == 0, "misclassification for case 4, dmethod=nocross"

        dl = deline.DLineate22(np.copy(td4),dmethod="nojag")
        dl.preprocess()
        dl.collect_break_points()

        p = [12.5,7.5]
        c = dl.d.classify_point(p)
        assert c == 0, "misclassification for case 4, dmethod=nojag"

        p = [25,7.5]
        c = dl.d.classify_point(p)
        assert c == -1, "misclassification for case 4, dmethod=nojag"

    ## runtime is long 
    def test__DLineate22__classify_points__case_4(self):
        print("DLineat22 test 5")

        data = deline.test_dataset__DLineateMC_1()

        dl = deline.DLineate22(data,dmethod="nojag")
        dl.preprocess()
        dl.collect_break_points()
        
        c = 0 
        for x in data: 
            l = dl.d.classify_point(x[:2])
            l_ = x[2]
            if l == l_: c += 1 
            ##print("classifying ", l,l_)
        assert c == 1779, "got {}".format(c)

    def test__DLineate22__collect_break_points__AND__classify_point__case_1(self):

        print("DLineat22 test 6")

        data = deline.test_dataset__Dlineate22_1_v2()
        dl = deline.DLineate22(data)
        dl.preprocess()
        dl.collect_break_points()

        for q in dl.lpoints:
            c = dl.d.classify_point(q)
            assert c == 0
     
    def test__DLineate22__optimize_delineation__case_1(self):
        print("DLineat22 test 7")

        data = deline.test_dataset__Dlineate22_1_v3(500,[0.000001,1.])
        dl = deline.DLineate22(data)
        dl.preprocess()
        dl.collect_break_points()
        s = dl.optimize_delineation()
        assert s > 0, "optimization should reduce classification error!"
    
    def test__DLineate22__full_process__case_1(self):
        print("DLineat22 test 8")

        xyl = np.array([[5.,0.,0],\
                [7.5,15.,0],\
                [20.,-5.,0],\
                [1,15,1],\
                [3,4,2],\
                [17,15,2]])

        dl = deline.DLineate22(xyl,dmethod="nocross")
        q = dl.full_process()
        X = xyl[:,:2]
        L = xyl[:,2]
        results = [] 
        for d,l in zip(X,L): 
            l_ = dl.d.classify_point(d)
            #print("got {} actual {}".format(l_,l))
            results.append(l_) 

        assert results == [-1, -1, -1, 1, -1, -1]

    def test__DLineate22__full_process__case_2(self): 

        xyl = np.array([[0.,0.,0],\
                [10,0,0],\
                [0,10,0],\
                [10,10,0],\
                [5,5,1],\
                [15,15,2]])

        dl = deline.DLineate22(xyl,dmethod="nocross",target_min_label=False)
        q = dl.full_process()
        assert q[:3] == ([0, 1, 2, 3, 4], Counter({0: 4, 1: 1}), 4)

        X = xyl[:,:2]
        L = xyl[:,2]

        results = [] 
        for d,l in zip(X,L): 
            l_ = dl.d.classify_point(d)
            print("got {} actual {}".format(l_,l))
            results.append(l_) 

        assert results == [0,0,0,0,0,-1]

if __name__ == '__main__':
    unittest.main()